#!PERL
#
# 
# Copyright (c) 1993, 1994 The RARE Association 
# 
# All Rights Reserved
# 
# Permission to use, copy, modify, and distribute this software and its 
# documentation for any purpose and without fee is hereby granted, 
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in 
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.  
# 
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#
#
#       $RCSfile: prtraceroute.pl,v $
#       $Revision: 1.9 $
#       $Author: david $
#       $Date: 1996/07/11 12:37:07 $
#
#
# This program will do a normal traceroute, and will add some AS info to
# the output, and will compare the real output with what is in the
# database concerning routing policies.
# 
# This version has a built in RIPE whois client, just to avoid compatibility
# problems. The default host to query is whois.ripe.net, if you have a RIPE
# whois server closer to home, change the variable WHOISHOST in the Makefile.
# Please note that a query to a non RIPE whois server makes no sense
# whatsoever, besides you may get some nice errors because of special flags
# used.
#
require "getopts.pl";

$whoishost = "WHOISHOST";
$traceroutecmd = "TRACEROUTE";
$hostcmd = "HOSTCMD";
$keepopen = "KEEPOPEN";
$fast = "FAST";
$version = "VERSION";
$port = PORT;
$AF_INET = 2;
$SOCK_STREAM = SOCKSTREAM;

# whois.pl - open whois connection, and returns socket descriptor.
# 
# Local whois client, to avoid forks and compatibility problems
# It basically is a fully-fledged RIPE version of the whois client.
# It leaves the socket descriptor open, the routine reading from the
# socket is supposed to close the socket again. Besides, the server
# will most probably close the connection, once all data has been output


sub whois {

    local($string) = @_;

    # print STDERR "QUERY: $string\n";

    if (! -S S) {
        $sockaddr = 'S n a4 x8';
        local($name, $aliases, $type, $len, $thataddr) = gethostbyname($whoishost);
        $that = pack($sockaddr, $AF_INET, $port, $thataddr);
        socket(S, $AF_INET, $SOCK_STREAM, $proto) || die "socket failed\n";
        if ( connect(S, $that) ) {
            if ( $opt_d) {
                printf "** connecting to whois host %s\n", $whoishost;
            }
        } else { 
            die "Connect to whois host $whoishost failed\n";
        }
        if ( $opt_d ) {
            printf "** connected ok\n";
        }
    }
    select(S); $| = 1; select(STDOUT);
    
    if($keepopen) {
        print S "-k ";
    }

    if ($fast) {
        print S "-F ";
    }

    if($version) {
        printf S "-Vpt%s ", $versionno;
    }
    
    print STDERR "query: -r ", $string, "\n" if ($opt_d);
    
    print S "-r ", $string, "\n";
    
    return S;
    
}

1;
#	$RCSfile: prtraceroute.pl,v $
#	$Revision: 1.9 $
#	$Author: david $
#	$Date: 1996/07/11 12:37:07 $
# 
# netinfo.pl - do all stuff needed to read as network information

#require "whois.pl";
#require "misc.pl";

# Now the return of this routine is rather ugly. Because the main routines
# only want to work with one type object, we kludge an inet-rtr inside
# a fake route object. At least, this is what it looks like to the rest
# of the programs. The object returned is basically a mix between the
# inet-rtr and route object. And only so that the other parts do not need
# to distuinguish between the two.

sub netinfo {

    local($host) = @_;

    print STDERR "netinfo - called for $host\n" if $opt_d;

    local($netint) = &quad2int($host);
    local($net)=$host;

    if (!$donenet{$net}) {
	print STDERR "netinfo for $net not yet done\n" if $opt_d;
        eval "\%net$netint = &getnetinfo(\"$net\");";
        $donenet{$net} = "net$netint";
    }
    eval "\*cur = \*$donenet{$net};";

    print STDERR "\$cur{\"rt\"} = $cur{\"rt\"}\n" if $opt_d;
    return *cur;
}


sub getnetinfo {

    local($net) = @_;
    local(%en) = ();
    local($gotpol) = 0;
    local($newline) = 0;
    local($skiptherest) = 0;
    local($rtrseen) = 0;
    local($rteseen) = 0;

    # first try to find an inetrtr!

    &whois("-a -T ir -L $net/32");

    while (1) {

        print STDERR "start loop ($newline)\n" if ($opt_d);
        
        $_=<S>;
        
	chop;
	
	if (/^\s*$/) {
	   print STDERR "newline\n" if ($opt_d);
	   $newline++;
	   last if ($newline>=2);
	   next;
	}
	else {
	   $newline=0;
	}

	print STDERR "NETINFO saw: $_\n" if ($opt_V);

	# Kludgerdekludge router object it into a route object ....

	if ((!$rtrseen) && (/^\inet\-rtr:\s+(.*)$/ || /^\*ir:\s+(.*)$/)) {
	    $rtrseen = 1;
	    $en{"rt"} = $1;
	    $gotpol = 1;

	    while(1) {
	        
	        $_=<S>;

		chop;

		print STDERR "ir: saw: $_\n" if ($opt_V);

		if (/^\s*$/) {
		   $newline++;
		   last;
		}
		else {
		   $newline=0;
		}

	        # More kludge, we are in an inet-rtr object
	        if (/^localas:\s+(.*)$/ || /^\*la:\s+(.*)$/) {
	            $en{"or"} = $1;

	            # print "origin: $1\n";
	            next;
	        }

	        # we also want to keep a list of peers and interface addresses
	        # for this inet-rtr. Fake again into a route object
	        if (/^ifaddr:\s*(\S+)\s+(\S+)$/ || /^\*if:\s*(\S+)\s+(\S+)$/) {
	            $en{"if"} .= "\n" if $en{"if"};
	            $en{"if"} .= "$1 $2";
	            next;
	        }

	        if (/^peer:\s*(\S+)\s+(AS\S+).*$/ || /^\*pe:\s*(\S+)\s+(AS\S+).*$/) {
	            $en{"pe"} .= "\n" if $en{"pe"};
	            $en{"pe"} .= "$1 $2";
	            next;
	        }
            
            }
            
	}
	elsif ($_=~ /^\s*\[\%\#\*]$/) {
	   $newline=0;
	   next;
	}
	else { #skip object

	    while(1) {
		
		$_=<S>;
		
		chop;

		print STDERR "skip ($newline): $_\n" if ($opt_V);

                if (/^\s*$/) {
	           $newline++;
	           last;
	        } 
	        else {
	           $newline=0;
	        }
	        
	    }
	    
        }
        
        last if ($newline>=2);
        
    }

    if (!$rtrseen) {

       $newline=0;
       
       #
       # trick to get other then 'ir' objects for non-split databases
       # /32 routes don't exist and it makes the queries even faster ;-)
       
       local($network)=$net;
       $network=~ s/\.(\d+)\s*$/\./;
       if ($1 % 2) {
          $network.=$1-1;
       }
       else {
          $network=$net;
       }
       
       &whois("-a -T rt -L $network/31");

       while (1) {
          
          print STDERR "start loop ($newline)\n" if ($opt_d);
        
          $_=<S>;
        
	  chop;
	
	  if (/^\s*$/) {
	     print STDERR "newline\n" if ($opt_d);
	     $newline++;
	     last if ($newline>=2);
	     next;
	  }
	  else {
	     $newline=0;
	  }
          
	  print STDERR "NETINFO rt: $_\n" if ($opt_V);

          if ((!$rteseen) && (!$rtrseen) && (!/\/[0-5]\s*$/) && (/^route:\s+(.*)$/||/^\*rt:\s+(.*)$/)) {

	     $en{"rt"}=$1;

	     # print "object ($_): ", $en{"rt"}, "\n";

	     while(1) {
		chop;
		
		$_=<S>;

		# print STDERR "rt: saw: $_\n";
                
                if (/^\s*$/) {
	           $newline++;
	           last;
	        } 
	        else {
	           $newline=0;
	        }
                
	        if (/^origin:\s+(.*)$/ || /^\*or:\s+(.*)/) {

	            if ($1 eq "AS0") {

	               while(1) {

                          $_=<S>;
                          
	                  # print STDERR "skip: $_\n";
                          if (/^\s*$/) {
	                     $newline++;
	                     last;
	                  } 
	                  else {
	                     $newline=0;
	                  }


		       }
		       
		       last if ($newline);

	            }
	            else {

	               $rteseen=1;

	               $en{"or"} = $1;
	               $gotpol = 1;

	               # print "origin: $1\n";

	               next;

	            }
	        }

                last if ($newline);
                
	        if (/^comm-list:\s+(.*)$/ || /^\*cl:\s+(.*)/) {
	            $en{"cl"} .= " " if $en{"cl"};
	            $en{"cl"} .= $1;
	            $gotpol = 1;
	            next;
	        }
	        if (/^remarks:\s+ias\-int:\s+(\S+)\s+(\S+)/ ||
	            /^\*rm:\s+ias\-int:\s+(\S+)\s+(\S+)/) {
	            $en{$1} = $2;
	            $dmz{$en{"rt"}} = 1;
            	    $dmzas{"$en{\"rt\"}%$2"} = 1;
            	    $dmzas{"$en{\"rt\"}%$en{\"or\"}"} = 1;
	            $gotpol = 1;
	            next;
	        }
	        
	        
             }
	  }
	  elsif ($_=~ /^\s*[\#\%\*]$/) {
	     $newline=0;
	     next;
	  }
	  else { #skip object

	     while(1) {
	     
	        $_=<S>;
	        
		chop;

		print STDERR "skip: $_\n" if ($opt_d);
                
                if (/^\s*$/) {
	           $newline++;
	           last;
	        } 
	        else {
	           $newline=0;
	        }
                
	     }
	 
	 }    
          
       }
       
    }

    return %en;

}

1;
#	$RCSfile: prtraceroute.pl,v $
#	$Revision: 1.9 $
#	$Author: david $
#	$Date: 1996/07/11 12:37:07 $
#
# asinfo.pl - do all stuff needed to read as information

# require "whois.pl";

sub asinfo {

    local($as) = @_;

    if (!$doneas{$as}) {
        eval "\%AS$as = &getinfo(\"$as\");";
        $doneas{$as} = "AS$as";
    }
    eval "\*cur = \*$doneas{$as};";

    return *cur;
}

sub getinfo {
    local($as) = @_;

    local(%en) = ();
    local($gotpol) = 0;
    local($newline) = 0;
    local($skiptherest) = 0;

    &whois("-a -T an $as");

    while (1) {

        $_=<S>;
        
        print STDERR "AS INFO Read: $_" if ($opt_d);
	chop;
	
	if (/^\s*$/) {
	   $newline++;
	   last if ($newline>=2);
	} 
	else {
	   $newline=0;
	}

	if ($skiptherest) {
#	    print STDERR "Ignoring: $_\n";
	    next;
	}

	if (/^aut-num:\s+(.*)$/ || /^\*an:\s+(.*)/) {
#	    $en{"an"} .= "\n" if $en{"an"};
	    $en{"an"} = $1 if !$en{"an"};
	    next;
	}

	if (/^default:\s+(.*)$/ || /^\*df:\s+(.*)/) {
	    $en{"df"} .= "\n" if $en{"df"};
	    $en{"df"} .= $1;
	    $gotpol = 1;
	    next;
	}

	if (/^as\-in:\s+(.*)$/ || /^\*ai:\s+(.*)/) {
	    ($peer, $weight, $policy) = split(/\s+/, $1, 3);
	    if (($peer eq $prevpeer) && ($weight eq $prevweight)) {
		$en{"ai"} .= " $policy";
	    } else {
		$en{"ai"} .= "\n" if $en{"ai"};
		$en{"ai"} .= "$peer $weight $policy";
	    }
	    $prevpeer = $peer;
	    $prevweight = $weight;
	    $gotpol = 1;
	    next;
	}

	if (/^as\-out:\s+(.*)$/ || /^\*ao:\s+(.*)/) {
	    $en{"ao"} .= "\n" if $en{"ao"};
            $en{"ao"} .= $1;
	    $gotpol = 1;
            next;
        }

	if (/^descr:\s+(.*)$/ || /^\*de:\s+(.*)/) {
            $en{"de"} = $1 if !$en{"de"};
	    $descr{$en{"an"}} = $en{"de"};
            next;
        }

	if (/^\s*$/ && $gotpol) {
	    $skiptherest = 1;
	    next;
	}
    }

    $en{"de"} = "??" if !$en{"de"};

    return %en;
}

1;
sub expandmacro {
    local($macro) = @_;

    local($result)="";
    local($newline) = 0;

    local(@toobeexpanded)=();

    # print STDERR "expand macro: -$macro-\n" if $opt_d;

    if ($asmacroknown{$macro}) {
        # print STDERR "expand macro, macro already known\n" if $opt_d;
	return $asmacroknown{$macro};
    }

    &whois("-a -T am $macro");

    while (1) {
        
        $_=<S>;
        
        if (/^\s*$/) {
	   $newline++;
	   last if ($newline==2);
	} 
	else {
	   $newline=0;
	}
#
        # print STDERR "MACRO Read: $_";

	if ((!$newline) && (/^as\-list:\s+(.*)$|^\*al:\s+(.*)$/)) {

	    local($value) = $1;

	    $value = $2 unless $value;

	    foreach $value (split(/\s+/, $value)) {

		if ($result) {
		    $result .= " OR ";
		}

		if ($value=~ /^\s*AS\-/) {

		    if ($asmacroknown{$macro}) {
		       $result.=$asmacroknown{$macro};
		    }
		    else {
		       $result.=$value;
		       push(@toobeexpanded,$value);
		    }

		}
		else {
		    $result.=$value;
		}

	    }
	    next;
	}
    }

    foreach $macro (@toobeexpanded) {
       local($regularmacro)=$macro;
       $regularmacro=~ s/([^\w\s])/\\\1/g;
       $result=~ s/$regularmacro/&expandmacro($macro)/ge;
    }

    $result=~ s/\s+/ /g;
    $result=~ s/^ //g;
    $result=~ s/ $//g;

    $asmacroknown{$macro} = "( $result ) ";

    # print STDERR "Total for $macro is ( $result )\n" if $opt_d;

    return "( $result ) ";

}



#
# quad2int uses * and / because perl5 seems to mess up bit shifts
#

sub quad2int {
    local($quad)=@_;
    local($result);

    # print STDERR "quad: -$quad-\n";
    
    if ($quad=~ /^\s*\*?\s*$/) {
       return ("");
    }
    else {
       if ($quad=~ /^0*(1?\d?\d|2[0-4]\d|25[0-5])\.0*(1?\d?\d|2[0-4]\d|25[0-5])\.0*(1?\d?\d|2[0-4]\d|25[0-5])\.0*(1?\d?\d|2[0-4]\d|25[0-5])$/) {
          $result=$1;
          $result*=256;
       
          $result+=$2;
          $result*=256;
       
          $result+=$3;
          $result*=256;
       
          $result+=$4;
          
          return ($result);
       }
       else {
          return -1;
       }
    }
}

1;
#       $RCSfile: prtraceroute.pl,v $
#       $Revision: 1.9 $
#       $Author: david $
#       $Date: 1996/07/11 12:37:07 $
#
# isnexthop - takes a ripe-81 expression, and target values for
# AS and communities (AS in a scalar, communities in a normal array)

sub isnexthop {

    local(*ripe81expr, *targetAS, *targetCOM, *targetPrefix) = @_;
    local($KEYWORDS) = 'NOT|OR|AND|\(|\)|\{|\}';
    local($newexpr, $expr);
    local(@matches) = ($targetPrefix, @targetCOM, $targetAS, "ANY");

    # print STDERR "expr: -$ripe81expr-\nTargets: $targetAS, $targetPrefix, ", @targetCOM, "\n" if $opt_d;

    $ripe81expr =~ s:(\(|\)|\{|\}): \1 :g;

    local(@blah) = split(/\s+/, $ripe81expr);

    foreach $i (0..$#blah-1) {


        next if ($blah[$i]=~ /^\s*$/);
        
        if (($blah[$i]=~ /^\)|\}$/) && ($blah[$i+1]!~ /^$KEYWORDS$/)) {
           $blah[$i].=" OR";
           next;
        }
        
        next if $blah[$i] =~ /^$KEYWORDS$/;
        next if $blah[$i+1] =~ /^$KEYWORDS$/;
        
        $blah[$i].= " OR";
    }
    for $i (0..$#blah) {
	if ($blah[$i] =~ /^AS\-/) {
	   #print STDERR "expand macro: $blah[$i]\n";
           splice(@blah, $i, 1, &expandmacro($blah[$i]));
           #print STDERR "expand macro done\n";
	}
        $newexpr.=$blah[$i]." "; 
    }

    print STDERR "evaluating $newexpr now\n" if $opt_d;
    
    foreach $i (split(/\s+/, $newexpr)) {
        
        # print STDERR "$i*";
	
	next if ($i=~ /^\s*$/);

	if ($i eq "NOT") {
	    $expr .= "!";
	    next;
	}
        if ($i eq "AND") {
            $expr .= "&";
            next;
        }
        if ($i eq "OR") {
            $expr .= "|";
            next;
        }
        if ($i eq "(") {
            $expr .= "(";
            next;
        }
        if ($i eq ")") {
            $expr .= ")";
            next;
        }
	if (($i eq "{")) {
	    $expr .= "(";
	    next;
	    }
	if ($i eq "}") {
	    $expr .= ")";
	    next;
	}
	if ($i eq ",") {
	    next;
	}

	foreach $j (@matches) {
	    if ($i eq $j) {
		$expr .= "1";
		$found = 1;
		last;
	    }
	}
	if (!$found) {
	    $expr .= "0";
	}
	$found = 0;
    }
    
    # do some corrections:
    
    $expr=~ s/(\)|\})(\d)/\1\|\2/g;
    $expr=~ s/(\d)(\(|\{)/\1\|\2/g;
    
    # print STDERR "\nexpr: ", $expr, "\n\n";
    local($result) = eval "$expr;";
    # print STDERR "expr done: -$result-\n";
            
#    print "$ripe81expr -> ", $result ? "Y" : "N", "\n";
    return $result;
}

1;
    
#	$RCSfile: prtraceroute.pl,v $
#	$Revision: 1.9 $
#	$Author: david $
#	$Date: 1996/07/11 12:37:07 $
# 
# nexthops.pl - find all nexthops, takes as input an AS number and a 
# target host address. Looks up info for these two if needed. Returns
# a string with all nexthops space weight, seperated by %.

# require "asinfo.pl";
# require "netinfo.pl";
# require "isnexthop.pl";

sub numerically { $a <=> $b ; }

sub nexthops {

    local($as, $target) = @_;
    local(@nexthop) = ();
    local($havepolicy)=0;

    print STDERR "nexthops - called with ($as, $target)\n" if $opt_d;

    *asen = &asinfo($as);

    if ($target !~ /^AS/) {
	*neten = &netinfo($target);
    } else {
	# This is not needed, but need side effect to set description
	*notused = &asinfo($target);
	$neten{"or"} = $target;
    }

    local($targetas) = $neten{"or"};
    local(@targetcom) = split(/\s+/, $neten{"cl"});
    local($targetprefix) = $neten{"rt"};

    print STDERR "nexthops got all main info, looking at policy now\n" if $opt_d;
    
    foreach (split(/\n/, $asen{"ai"})) {
	local($peer, $weight, $policy) = split(/\s+/, $_, 3);
	if (&isnexthop(*policy, *targetas, *targetcom, *targetprefix)) {
	    $nexthop[$index++] = "$weight $peer";
	}
	$havepolicy=1;
    }
    
    print STDERR "nexthops - as seen\n" if $opt_d;
    
    if ($asen{"df"}) {
	foreach (split(/\n/, $asen{"df"})) {
	    local($peer, $weight) = split(/\s+/, $_);
	    $weight+=50000;
	    $nexthop[$index++] = "$weight $peer (D)";
	}
	$havepolicy=1;
    }

    
    return ($havepolicy, sort numerically @nexthop);

}

1;

# &Nexthops($ARGV[0], $ARGV[1]);

#print "next hops for $ARGV[1] from $ARGV[0] are:\n";
#for $i (sort numerically @nexthop) {
#    print "$i\n";
#}




# This is a associative array used for netinfo. Must be used globally.

%neti = ();
%asi = ();

#

#
&Getopts('vdlm:q:w:h:p:g:KFV');
#
$usage="Usage: $0 [-v] [-l] host\nOther: [-d] [-m maxhops] [-q nqueries] [-w waittime] [-g gateway]\n       [-h whoishost] [-p port]";

if ($#ARGV != 0) {
    print STDERR "$usage\n";
    exit 1;
}

if ($ENV{"PRIDEHOST"}) {
	$whoishost = &getfqdn($ENV{"PRIDEHOST"});
}
if ($opt_h) {
    $whoishost = &getfqdn($opt_h);
}
if ($ENV{"PRIDEPORT"}) {
    $port = $ENV{"PRIDEPORT"};
}
if ($opt_p) {
    $port = $opt_p;
}

$traceroutecmd=~ /^\s*(\S+)(\s|$)/;

if ($traceroutecmd=~ /\//) {
   
   if (! -f $1) {
   
      print STDERR "Cannot find traceroute ($1) command\n";
   
      exit 1;
   }

   if (! -x $1) {
   
      print STDERR "traceroute ($1) command is not an executable file\n";
   
      exit 1;
   }

}

if ($opt_m) {
    $maxhops = $opt_m;
    $traceroutecmd = $traceroutecmd." -m ".$maxhops;
}

if ($opt_q) {
    $queries = $opt_q;
    $traceroutecmd = $traceroutecmd." -q ".$queries;
}

if ($opt_w) {
    $waittime =  $opt_w;
    $traceroutecmd = $traceroutecmd." -w ".$waittime;
}

if ($opt_g) {
    $viahost =  &getfqdn($opt_g);
    $traceroutecmd = $traceroutecmd." -g "."$opt_g";
}
if ($keepopen =~ /^TRUE$/i) {
    $keepopen = 1;
} else {
    $keepopen = 0;
}
if ($opt_K) {
    $keepopen = 0;
}
if ($fast =~ /^TRUE$/i) {
    $fast = 1;
} else {
    $fast = 0;
}
if ($opt_F) {
    $fast = 0;
}
if ($version=~ /^TRUE$/i ) {
    $versionno = '$Revision: 1.9 $';  # no need to fool emacs anymore,
                                      # since I use joe ;-)
                                      # the quote trick didn't work for perl5 
    $versionno =~ s/\$Revision\: //;
    $versionno =~ s/ \$$//;
    $version = 1;
} else {
    $version = 0;
}
if ($opt_V) {
    $version = 0;
}


if ($opt_d) {
    printf "** %s use keepopen feature\n", 
    $keepopen ? "will" : "will not" ;
}


# Needed for sorting things numerically in stead of alphabetically

sub numerically { $a <=> $b ; }


#
# Return the UTC time for verbose output
#
sub gettime {

    local($s, $m, $h, $md, $mo, $y, $wd, $yd, $is) = gmtime(time);
    
    $thismonth = ("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec")[$mo];

    $s = "0".$s unless $s > 9;
    $m = "0".$m unless $m > 9;
    $h = "0".$h unless $h > 9;

    $date = sprintf("%3s %2d %2s:%2s:%2s UTC", 
                    $thismonth, $md, $h, $m , $s );
    return $date;

}
# 
# Return an IP address from a name.
#
sub getadr {
    local($host) = @_;
    local($name, $alias, $addrtype, $length, @addr) = gethostbyname($host);
    if (defined($addr[0])) {
	local($a, $b, $c, $d) = unpack('C4', $addr[0]);
	$addr = $a.".".$b.".".$c.".".$d;
	return($addr);
    } else {
	print STDERR "$0: unknown address for $host\n";
	exit 1;
    }
}
#
# Return the fully qualified domain name.
#
sub getfqdn {
    local($host) = @_;
#
# resolve an address first
#
    if($host =~ /^(\d+)+\.(\d+)\.(\d+)\.(\d+)$/) { 
	local($fqdntmp) = pack('C4', $1, $2, $3, $4);
	local($name, $alias, $addrtype, $length, @addr) = gethostbyaddr($fqdntmp,2);
	if (defined($name)) {
	    return($name);
	} else {
	    return($host);
	}
    }	
#
# okay then try a name
#	    
    ($name, $alias, $addrtype, $length, @addr) = gethostbyname($host);  
    	
    if (!defined($name)) {
	print STDERR "$0: unknown host $host\n";
	exit 1;
    }
    return($name);
}

if ($opt_d) {
    printf "** traceroute command will be \"%s\"\n", $traceroutecmd." ".$ARGV[0];
}

$destname = &getfqdn($ARGV[0]);

if (!open(TRACE, "$traceroutecmd $destname 2>&1 |")) {
   
   print STDERR "couldn't start: $traceroutecmd\n";
   print STDERR "please check if it's in you path or has a wrong definition\n";
   
   exit 1;
}

$nrofhops=0;

while (<TRACE>) {
    
    # strip leading spaces, trailing spaces, commas and stars (traceroute timeouts,
    # they mess up the ordering of fields, making it impossible to parse)
   
    s/^\s+//;
    s/\s+$//;
    s/\*//g;
    s/\,//g;
    s/\s+/ /g;

    # print "*** ", $_, "\n";

    # next if we have nothing left....
    
    next if (/^\s*\d*\s*$/);

    # traceroute found an error	
	
    if (/traceroute: /) {
	($nothing, $error) = split(/:/, $_);
	print STDERR "$0: $error\n";
	exit 1;
    }

    # First let's find out the AS for the destination network from the first line
    
    if (/^traceroute/) {		
	
	if (!(($hostname,$destnum)=/\s(\S+)\s*[^\d\s]*(\d+\.\d+\.\d+\.\d+)[^\d\s]/)) {
	   
	   print STDERR "You are possibly using an incompatible version of traceroute\n";
	   print STDERR "Please check the documentation for a compatible version\n";
	   
	   exit 1;

	}

	*neti = &netinfo($destnum);
	$destnet = $neti{"rt"};
	$das = $neti{"or"};

	if($opt_g) {
	    if (&quad2int($opt_g)>=0) {
		$viaaddr = $opt_g;
	    }
	    else {
		$viaaddr = &getadr($viahost);
		# print STDERR "via: $viaaddr\n";
	    }
	    *neti = &netinfo($viaaddr);
	    $viaas = $neti{"or"};
	}

	if (!$das) {
	    print STDERR "** WARNING ** Destination AS unknown ".
		"for $hostname \($destnum\)\n";
	    $setnopolicy = 1;
	}

	if ($opt_g && !$viaas) {
	    print STDERR "** WARNING ** Gateway AS unknown ".
		    "for $viahost \($viaaddr\)\n";
		$setnopolicy = 1;
	}

	if ($setnopolicy) {
	    print STDERR "** WARNING ** Policy information is not ".
		"possible - setting to \"?\"\n\n";
	}
  				
# Try to get the originating AS and other local information.

	local($tmph) = `$hostcmd`;
	chop($tmph);
	$localhost = &getfqdn($tmph);
	$localaddr = &getadr($localhost);

	*neti = &netinfo($localaddr);
	$localas = $neti{"or"};
	if ($neti{$localaddr}) {
            $localas = $neti{$localaddr};
        }
	if ($localas) {
	    $aspath[++$pathindex] = $localas;
	}
	if($opt_g) {
	    $destas = $viaas;
	} else {
	    $destas = $das;
	}
	if ($opt_v) {
	    $time = &gettime();
	    if (!$localas) {
		$localas = "???";
	    }
	    printf "traceroute with AS and policy additions [%s]\n\n", $time;
            printf "\t\tfrom %6s %s \(%s\)\n",
	    $localas ? $localas : "???", $localhost, $localaddr;
	    if($opt_g) {
		printf "\t\tvia  %6s %s \(%s\)\n", 
		$viaas ? $viaas : "???" , &getfqdn($viahost), $viaaddr;
	    }
	    printf "\t\tto   %6s %s \(%s\)\n\n", 
	    $das ? $das : "???", &getfqdn($hostname), $destnum;
	}
	else {
	    printf "traceroute to %s \(%s\) with AS and policy additions\n\n", $hostname, $destnum;
	}
	next;
    }

    # Here come all the hops
    
    if (/^\d+\s/) {
    
        $nrofhops++;
        
	# Get the IP address out of the traceroute output

	($hop,$hostname,$netnum,$rest)= /^(\d+)\s+(\S+)\s+[^\d]*(\d+\.\d+\.\d+\.\d+)[^\d]*\s+(\S.*)$/;
	
        # Do something special for first hop. You could be in another AS 
	
	if ($hop == 1) {
	    $prevas=$localas;
	}
	else {
	    $prevas=$as;
	}

        # Get the as number for this IP address:

	*neti = &netinfo($netnum);
	$as = $neti{"or"};

	if ($neti{$netnum}) {
            $as = $neti{$netnum};
        }

	$lookupnet = $neti{"rt"};
    
        # See if we figure out what kind of routing decision this was according
        # to policy in the database

	if (!$as) {
	    $as = "???";
	}

        $type="";
        
	if (($as eq $prevas) && ($as !~ /\?/)) {
	    $type = "I";
	}
	elsif ($as =~ /\?/) {
	    $type = "?";
	}
	elsif ($prevas =~ /\?/) {
	    $type = "?";
	}
	else {
	    local($havepolicy);
	    ($havepolicy, @nexthops)= &nexthops($prevas, $destnum);
	    print STDERR "nexthops returned\n" if $opt_d;
	    $level = 0;
	    $cost = 0;
	    $prevcost = 0;
	    for $i (0..$#nexthops) {
		print "** looking at nexthop $nexthops[$i]\n" if $opt_d;
		$prevcost = $cost;
		$cost = sprintf("%d", $nexthops[$i]);
		if ($prevcost - $cost != 0) {
		    $level++;
		}
		if ($nexthops[$i] =~ /$as/) {
		    print STDERR "$nexthops[$i] matches $as\n" if $opt_d;
		    if ($nexthops[$i] =~ /\(D\)/) {	
			$type = "D$level";
		    } else {
			$type = "E$level";
		    }
		}
		last if $type;
	    }
	    # This is the new next hop stuff, that can deal with
	    # ifaddr and peer, if that fails, we fall back to old
	    # mode ias-int ...

	    if (!$type) {
		if (!@nexthops) {
		   $type = "?";
		}
		elsif ($havepolicy) {
		   $type="ERROR";
		}
		else {
		   $type="?";
		}
	    }
		    
# This is the possible next hop stuff, leave that for a bit later
	    
#	    if (!$type) {
#		if (!@nexthops) {
#		    $type = "?";
#		} else {
#		    ($tmp, $possas, $tmp) = split(/\s+/, $nexthops[0]);
#		    $possas =~ s/AS//;
#		    if ($dmzas{"$destnet%$prevas"} && 
#			($lookupnet eq $destnet)) {
#			$type = "C";
#		    } elsif ($dmzas{"$lookupnet%$possas"}) {
#			local($tmp) = $possas;
#			$tmp =~ s/AS//;
#			if($destas) {
#			    $aspath[++$pathindex] = "[?".$tmp."?]";
#			}
#			&asinfo($tmp);
#			$type = "?NH $possas";
#		    } else {
#			$type = "ERROR";
#		    }
#		}
#	    }
	}
	
	if($opt_g && ($as eq $viaas)) {
	    $as = $viaas;
	    $destas = $das;
	}
    }
    

# And print the output

    $rest =~ s/\(\S+\)//g;
    $icmpmess = $rest;
    $icmpmess =~ tr/ ![HPFS]*//cd;
    $icmpmess =~ s/\s+$//;
    $icmpmess =~ s/\s+/ /g;
    $icmpmess =~ s/^\s+//;
    
    # print "level: -$level- type: -$type-\n";
    
    if ($opt_l) {

#
# Try to do something sensible with this - not too easy as traceroute can
# give very unpredictable results
# This is only really a first stab for those wanting a possibilty to see RTT's
# May try better with this later.
#
	$rest =~ s/ \![HPFS]//g;
	$rest =~ s/ ms//g;
	$rest =~ s/(\S)\s*$/$1 ms/;
	printf "%2d %6s  %-30s %-18s %s %s%s\n", 
	$nrofhops, $as, substr($hostname,0,30), $netnum, 
	$setnopolicy ? "[?]" : "[".$type."]", $rest, 
        $icmpmess ? " $icmpmess" : "";
	$rest = "";
    } else {
	printf "%2d %6s  %-30s %-18s %s%s\n",
	$nrofhops, $as, substr($hostname,0,30), $netnum, 
	$setnopolicy ? "[?]" : "[".$type."]", 
	$icmpmess ? " $icmpmess" : "" ;
    }
    $type = "";
    $icmpmess = "";

# Update the path

    if ($as ne $prevas) {
	$aspath[++$pathindex] = "$as";
    }
    
}

# End of traceroute output

close(TRACE);

print STDERR "closed trace and whois output now\n" if $opt_d;
# Print path, legend and whois stats

if ($#aspath >= 1) {
    print "\nAS Path followed: ", join(" ",@aspath), "\n\n";
    for $i (1..$#aspath) {
	if ($aspath[$i] ne "???") {
	    if ($aspath[$i] =~ m/\[.(\S+).]/) {
		$aspath[$i] = $1;
	    }
	    if (!$done{$aspath[$i]}) {
		*asi = &asinfo($aspath[$i]);
		printf "%6s = %s\n", 
		$aspath[$i], $asi{"de"},
		$done{$aspath[$i]} = 1;
	    }
	}
    }
}
close(S);

# 
# Give some debug stats for those who like this sort of thing
#
if ($opt_d) {
    printf "\n** (needed %d whois queries - %d network, %d asnumbers)\n", 
    $netwhoisq+$aswhoisq,$netwhoisq, $aswhoisq;
}
